/****************************************************************************
 ****************************************************************************

                        Datenkomprimierung

 ****************************************************************************
 ****************************************************************************

        QUELLDATEI                COMPRESS.C

        ABATRON AG                6221 RICKENBACH
        ENTWICKLER                Ph.Simon
        DATUM                     24.05.1995


 ****************************************************************************
 *      Beschreibung :                                                      *
 ****************************************************************************

 Der Komprimieralgorithmus basiert auf den LZ77-Komprimieralgorithmus von
 Lempel und Ziv mit einem gleitenden Fenster ueber vorher eingelesene Zeichen.
 Das gleitende Fenster hat eine Groesse von 4 kBytes.

 Um die Suche nach uebereinstimmenden Zeichenfolgen zu beschleunigen, wird fuer
 jedes ASCII-Zeichen aus dem Fenster eine einfach verkettete Liste aufgebaut.
 Geloeschte Zeichen werden am Kopf der Liste entfernt und neue Zeichen am Schwanz
 der Liste hinzugefuegt.

 Zeichen die keine oder eine Uebereinstimmung finden, werden unkodiert ausgegeben.
 Zeichenfolgen mit 2 bis 17 Uebereinstimmung werden als 16 Bit-Wert
 ( 4 Laengenbits  + 12 Positionsbits) ausgegeben.
 Um zu entscheiden, ob es sich um ein unkodiertes Datenbyte oder um ein Codewort
 handelt, Wird vor jeder Sequenz von 8 Elementen (Datenbyte oder Codewort) ein
 Byte mit der Type Information uebermittelt. Falls ein Bit gesetzt ist, heisst
 dies, dass es sich um ein Codewort handelt. Bit0 definiert den Typ des ersten
 Feldes nach dem Typebyte.

 Der komprimierte Datenstrom ist also wie folgt ausgebaut:

 <Type><Data0><Data1><Data2><Data3><Data4><Data5><Data6><Data7>
 <Type><Data0><Data1><Data2><Data3><Data4><Data5><Data6><Data7>
                           ..............
 <Type><Data0><Data1><Data2><Data3><Data4><Data5><Data6><Data7>

 Das <Type> Byte definiert die Typen der nachfolgenden <Data> Felder.
 Jedes Bit im <Type> entspricht einem Daten Feld. Bit0 entspricht
 Datafeld 0, Bit7 entspricht Datafeld 7.

 Bit = 0: Datafeld ist ein unkomprimiertes Byte (8bit)
 Bit = 1: Datafeld ist eine Kodierte Sequenz (16bit)
          <len(4bit)><index(12bit)>
       len = Sequenz Length - 2 (0..15 <=> 2..17)

 Mit dieser Art der Kodierung befinden sich alle Felder immer
 an Bytegrenzen. Aufwendige Bitschieberei entfaellt.


 ****************************************************************************/

/***** INCLUDE-DATEIEN ******************************************************/

#include <stddef.h>
#include "compress.h"

/***** CONSTANTS ***********************************************************/

#define INDEX_BITS        12
#define MATCH_BITS        4
#define TABLE_SIZE        ( 1 << INDEX_BITS )
#define MAX_MATCH         ( ( 1 << MATCH_BITS) + 1 )
#define HEAD_SIZE         256

/***** TYPEDEFS *************************************************************/

typedef struct Node*    NodePtrT;
typedef struct Node{
  NodePtrT      succ;
  NodePtrT      pred;
  NodePtrT      next;
  int           index;
  unsigned char value;
} NodeT;

/***** GLOBALE VARIABELN ****************************************************/

static NodeT    nodeTable[ TABLE_SIZE ];
static NodeT    tailTable[ HEAD_SIZE ];
static NodeT    headTable[ HEAD_SIZE ];
static NodePtrT tablePos;


/****************************************************************************
 ****************************************************************************

    LZA_Compress :

    Komprimiert einen Block ASCII-Zeichen.

    INPUT  : inCount    Anzahl zu komprimierende Zeichen
             inData     Zeiger auf den zu komprimierenden Zeichenblock
    OUTPUT : outData    Zeiger auf den Buffer mit den komprimierten Daten
             RETURN     Laenge des komprimierten Buffer in Bytes.

 ****************************************************************************/

int LZA_Compress(int inCount, unsigned char *inData, unsigned char *outData)
{
  unsigned char *inDataStartPtr;
  unsigned char *inDataNextPtr;
  unsigned char *outDataPtr;
  unsigned char *typeBytePtr;    /* points to current type field */
  unsigned char  newChar;
  unsigned char  typeByte;
  unsigned char  typeMask;
  NodePtrT       node;
  NodePtrT       next;
  int            hitLength;
  int            hitPos;
  int            maxMatch;
  int            i;

  /* init local variables */
  inDataStartPtr = inData;
  typeBytePtr    = outData;
  outDataPtr     = outData + 1;
  typeMask       = 1;
  typeByte       = 0;

  /* compress loop */
  while (inCount > 0)
  {
    /* find max matching sequence in window */
    maxMatch  = (inCount > MAX_MATCH) ? MAX_MATCH : inCount;
    hitLength = 1;
    node = headTable[*inDataStartPtr].succ;
    if ((node != NULL) && (maxMatch > 1)) {
      while ((node->succ != NULL) && (hitLength < maxMatch)) {
        inDataNextPtr = inDataStartPtr+1;
        next = node->next;
        for (i = 1; i < maxMatch; i++) {
          if (next->value != *inDataNextPtr++) break;
          next = next->next;
        } /* for */
        if (i > hitLength) {
          hitLength = i;
          hitPos = node->index;
        } /* if */
        node = node->succ; /* next node */
      } /* while */
    } /* if */


    /* store next field in output buffer */
    if (hitLength <= 1)
    {
      *outDataPtr++ = *inDataStartPtr;
    } /* if */
    else
    {
      *outDataPtr++ = (unsigned char)(((hitLength - 2) << (INDEX_BITS - 8)) + (hitPos >> 8));
      *outDataPtr++ = (unsigned char)hitPos;
      typeByte |= typeMask;
    } /* else */
    typeMask <<= 1;

    /* store last type byte in output buffer if block is full */
    if (typeMask == 0)
    {
      *typeBytePtr = typeByte;
      typeBytePtr = outDataPtr;
      outDataPtr++;
      typeByte = 0;
      typeMask = 1;
    } /* if */


    /* remove old values, insert new ones */
    for (i = 0; i < hitLength; i++)
    {
      /* remove if not empty */
      if (tablePos->succ != NULL) {
        tablePos->pred->succ = tablePos->succ;
      } /* if */

      /* insert in new char list */
      newChar = *inDataStartPtr++;
      node = &headTable[newChar];
      node->succ->pred = tablePos;
      tablePos->succ   = node->succ;
      tablePos->pred   = node;
      tablePos->value  = newChar;
      node->succ       = tablePos;

      tablePos = tablePos->next;
    } /* for */
    inCount -= hitLength;
  } /* while */

  /* store last zype byte or remove empty type byte */
  if (typeMask != 1) *typeBytePtr = typeByte;
  else               outDataPtr--;

  /* return number of bytes in compressed buffer */
  return (int)(outDataPtr - outData);

} /* LZA_Compress */


/****************************************************************************
 ****************************************************************************

    LZA_CompressInit :

    Initialisiert die Datenkomprimierungs Tabellen.

    INPUT  : -
    OUTPUT : -

 ****************************************************************************/

void LZA_CompressInit(void)
{
  int       i;
  NodePtrT  node;
  NodePtrT  head;
  NodePtrT  tail;

  /* init node table */
  node = nodeTable;
  for (i = 0; i < TABLE_SIZE; i++) {
    node->succ  = NULL;
    node->value = 0;
    node->index = i;
    node->next  = node+1;
    node++;
  } /* for */
  node--;
  node->next = nodeTable;

  /* init linked lists */
  head = headTable;
  tail = tailTable;
  for (i = 0; i < HEAD_SIZE; i++) {
    tail->succ = NULL;
    head->succ = tail++;
    head++;
  } /* for */

  tablePos = nodeTable;

} /* LZA_CompressInit */

